from typing import Optional

import typer
from ruamel.yaml import YAML

from ibek.ioc_cmds.commands import ioc_cli
from ibek.runtime_cmds.commands import runtime_cli
from ibek.support_cmds.commands import support_cli

from ._version import __version__
from .globals import NaturalOrderGroup

cli = typer.Typer(cls=NaturalOrderGroup)

cli.add_typer(
    support_cli,
    name="support",
    help="Commands for building support modules during container build",
)
cli.add_typer(
    ioc_cli,
    name="ioc",
    help="Commands for building generic IOCs during container build",
)
cli.add_typer(
    runtime_cli,
    name="runtime",
    help="Commands for building IOC instance startup files at container runtime",
)

yaml = YAML()


def version_callback(value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()


@cli.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print the version of ibek and exit",
    )
):
    """IOC Builder for EPICS and Kubernetes

    Provides support for building generic EPICS IOC container images and for
    running IOC instances in a Kubernetes cluster.
    """


# test with:
#     pipenv run python -m ibek
if __name__ == "__main__":
    cli()



import sys

if sys.version_info < (3, 8):
    from importlib_metadata import version  # noqa
else:
    from importlib.metadata import version  # noqa

__version__ = version("ibek")
del version

__all__ = ["__version__"]



"""
A few global definitions
"""

import os
from pathlib import Path
from typing import Dict

from jinja2 import Template
from pydantic import BaseModel, ConfigDict
from typer.core import TyperGroup

from .utils import UTILS

# get the container paths from environment variables
EPICS_BASE = Path(os.getenv("EPICS_BASE", "/epics/epics-base"))
EPICS_ROOT = Path(os.getenv("EPICS_ROOT", "/epics/"))
IOC_FOLDER = Path(os.getenv("IOC", "/epics/ioc"))
SUPPORT = Path(os.getenv("SUPPORT", "/epics/support"))

# the global RELEASE file which lists all support modules
RELEASE = SUPPORT / "configure/RELEASE"
# a bash script to export the macros defined in RELEASE as environment vars
RELEASE_SH = SUPPORT / "configure/RELEASE.shell"
# global MODULES file used to determine order of build
MODULES = SUPPORT / "configure/MODULES"

# Folder containing templates for IOC src etc.
TEMPLATES = Path(__file__).parent / "templates"
IBEK_FILES = EPICS_ROOT / "ibek"

IOC_DBDS = SUPPORT / "configure/dbd_list"
IOC_LIBS = SUPPORT / "configure/lib_list"
RUNTIME_DEBS = SUPPORT / "configure/runtime_debs"


class BaseSettings(BaseModel):
    """A Base class for setting consistent Pydantic model configuration"""

    model_config = ConfigDict(
        extra="forbid",
    )


class NaturalOrderGroup(TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()


def render_with_utils(context: Dict, template_text: str) -> str:
    """
    Render a Jinja template with the global __utils__ object in the context
    """
    jinja_template = Template(template_text)
    return jinja_template.render(context, __utils__=UTILS)



from pathlib import Path

from pytest import fixture
from ruamel.yaml import YAML
from typer.testing import CliRunner

from ibek.__main__ import cli
from ibek.ioc import clear_entity_model_ids, make_entity_models
from ibek.support import Support

runner = CliRunner()


def run_cli(*args):
    result = runner.invoke(cli, [str(x) for x in args])
    if result.exception:
        raise result.exception
    assert result.exit_code == 0, result


def get_support(yaml_file: str) -> Support:
    """
    Get a support object from the sample YAML directory
    """
    # load from file
    d = YAML(typ="safe").load(yaml_file)
    # create a support object from that dict
    support = Support(**d)
    return support


@fixture
def templates():
    return Path(__file__).parent.parent / "src" / "ibek" / "templates"


@fixture
def samples():
    return Path(__file__).parent / "samples"


@fixture
def yaml_defs(samples):
    return samples / "yaml"


@fixture
def objects_classes(yaml_defs):
    # clear the entity classes to make sure there's nothing left
    clear_entity_model_ids()

    objects_support = get_support(yaml_defs / "objects.ibek.support.yaml")

    # make entity subclasses for everything defined in it
    namespace = make_entity_models(objects_support)

    # return the namespace so that callers have access to all of the
    # generated dataclasses
    return namespace


@fixture
def epics_support(ibek_defs):
    return get_support(ibek_defs / "_global", "epics.ibek.support.yaml")


@fixture
def epics_classes(epics_support):
    # clear the entity classes to make sure there's nothing left
    clear_entity_model_ids()

    # make entity subclasses for everything defined in it
    namespace = make_entity_models(epics_support)

    # return the namespace so that callers have access to all of the
    # generated dataclasses
    return namespace



"""
Functions for rendering lines in the boot script using Jinja2
"""

from typing import Callable, List, Optional, Union

from .globals import render_with_utils
from .ioc import IOC, Entity
from .support import Comment, Script, Text, When


class Render:
    """
    A class for generating snippets of startup script / EPICS DB
    by using Jinja to combine snippet templates from support module
    definition yaml with substitution values supplied in ioc entity yaml
    """

    def __init__(self: "Render"):
        self.once_done: List[str] = []

    def render_text(
        self, instance: Entity, text: str, when=When.every, suffix=""
    ) -> str:
        """
        Add in the next line of text, honouring the once flag which will
        only add the line once per IOC.

        Jinja rendering of values/args has already been done in Entity.__post_init__
        but we pass all strings though jinja again to render any other jinja
        in the IOC (e.g. database and function entries)

        once uses the name of the definition + suffix to track which lines
        have been rendered already. The suffix can be used where a given
        Entity has more than one element to render once (e.g. functions)
        """

        if when == When.first.value:
            name = instance.__definition__.name + suffix
            if name not in self.once_done:
                self.once_done.append(name)
            else:
                return ""
        elif when == When.last.value:
            raise NotImplementedError("When.last not yet implemented")

        # Render Jinja entries in the text
        result = render_with_utils(instance, text)  # type: ignore

        if result == "":
            return ""

        return result + "\n"

    def render_script(self, instance: Entity, script_items: Script) -> Optional[str]:
        script = ""

        for item in script_items:
            if isinstance(item, Comment):
                comments = "\n".join(["# " + line for line in item.value.split("\n")])
                script += self.render_text(
                    instance, comments, item.when, suffix="comment"
                )
            elif isinstance(item, Text):
                script += self.render_text(
                    instance, item.value, item.when, suffix="text"
                )

        return script

    def render_pre_ioc_init(self, instance: Entity) -> Optional[str]:
        """
        render the startup script by combining the jinja template from
        an entity with the arguments from an Entity
        """
        pre_init = instance.__definition__.pre_init
        return self.render_script(instance, pre_init)

    def render_post_ioc_init(self, instance: Entity) -> Optional[str]:
        """
        render the post-iocInit entries by combining the jinja template
        from an entity with the arguments from an Entity
        """
        post_init = instance.__definition__.post_init
        return self.render_script(instance, post_init)

    def render_environment_variables(self, instance: Entity) -> Optional[str]:
        """
        render the environment variable elements by combining the jinja template
        from an entity with the arguments from an Entity
        """
        variables = getattr(instance.__definition__, "env_vars")
        if not variables:
            return None

        env_var_txt = ""
        for variable in variables:
            # Substitute the name and value of the environment variable from args
            env_template = f"epicsEnvSet {variable.name} {variable.value}"
            env_var_txt += render_with_utils(instance, env_template)  # type: ignore
        return env_var_txt + "\n"

    def render_elements(
        self, ioc: IOC, method: Callable[[Entity], Union[str, None]]
    ) -> str:
        """
        Render elements of a given IOC instance based on calling the correct method
        """
        elements = ""
        for entity in ioc.entities:
            if entity.entity_enabled:
                element = method(entity)
                if element:
                    elements += element
        return elements

    def render_pre_ioc_init_elements(self, ioc: IOC) -> str:
        """
        Render all of the startup script entries for a given IOC instance
        """
        return self.render_elements(ioc, self.render_pre_ioc_init)

    def render_post_ioc_init_elements(self, ioc: IOC) -> str:
        """
        Render all of the post-iocInit elements for a given IOC instance
        """
        return self.render_elements(ioc, self.render_post_ioc_init)

    def render_environment_variable_elements(self, ioc: IOC) -> str:
        """
        Render all of the environment variable entries for a given IOC instance
        """
        return self.render_elements(ioc, self.render_environment_variables)



import re
import subprocess
from enum import Enum
from pathlib import Path
from shutil import rmtree
from typing import List, Optional

import typer

from ibek.ioc_cmds.assets import get_ioc_source

try:
    from git import Repo
except ImportError:
    pass  # Git Python is not needed for runtime (container build time only)

from typing_extensions import Annotated

from ibek.globals import (
    IBEK_FILES,
    IOC_DBDS,
    IOC_LIBS,
    RELEASE,
    RUNTIME_DEBS,
    SUPPORT,
    NaturalOrderGroup,
)
from ibek.support import Support
from ibek.support_cmds.checks import (
    add_macro,
    do_dependencies,
    verify_release_includes_local,
)
from ibek.support_cmds.files import (
    Arch,
    add_list_to_file,
    add_text_once,
    get_config_site_file,
)


class AptWhen(str, Enum):
    dev = "dev"
    run = "run"
    both = "both"


# find macro name and macro value in a RELEASE file
# only include values with at least one / (attempt to match filepaths only)
PARSE_MACROS = re.compile(r"^([A-Z_a-z0-9]*)\s*=\s*(.*/.*)$", flags=re.M)


support_cli = typer.Typer(cls=NaturalOrderGroup)


@support_cli.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def apt_install(
    ctx: typer.Context,
    debs: List[str] = typer.Argument(None, help="list of debian packages to install"),
    only: AptWhen = typer.Option(
        AptWhen.both, help="which container build stage to install in"
    ),
    runtime: bool = typer.Option(False, help="install list of runtime packages"),
):
    """
    Install debian packages into the container. If they have an http:// or https://
    prefix then they will be downloaded and installed from file.
    """
    temp = Path("/tmp")

    if (only is AptWhen.run) or (only is AptWhen.both):
        add_list_to_file(RUNTIME_DEBS, debs)
    if only is AptWhen.run:
        return

    if runtime and RUNTIME_DEBS.exists():
        debs += RUNTIME_DEBS.read_text().split()

    for i, pkg in enumerate(debs):
        if pkg.startswith("http://") or pkg.startswith("https://"):
            pkg_file = temp / pkg.split("/")[-1]
            subprocess.call(["wget", pkg, "-O", str(pkg_file)])
            debs[i] = str(pkg_file)

    command = (
        "apt-get update && apt-get upgrade -y && "
        "apt-get install -y --no-install-recommends "
        + " ".join(debs)
        + " ".join(ctx.args)
    )
    exit(subprocess.call(["bash", "-c", command]))


@support_cli.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def git_clone(
    ctx: typer.Context,
    repo_name: str = typer.Argument(..., help="repo to clone"),
    version: str = typer.Argument(..., help="tag to clone"),
    org: str = typer.Option(
        "https://github.com/epics-modules/", help="repo organization URL"
    ),
    force: bool = typer.Option(False, help="overwrite existing clone"),
):
    """
    clone a support module from a remote repository
    """
    url = org + repo_name
    location = SUPPORT / repo_name
    if location.exists() and not force:
        print(f"skipping {location}, already cloned")
        return
    else:
        rmtree(location, ignore_errors=True)

    Repo.clone_from(
        url, SUPPORT / repo_name, branch=version, depth=1, multi_options=ctx.args
    )


@support_cli.command()
def register(
    name: str = typer.Argument(..., help="the name of the support module"),
    path: Annotated[Optional[Path], typer.Option(help="path to support module")] = None,
    macro: Optional[str] = typer.Option(None, help="Macro name for the module"),
):
    """
    prepare the configure RELEASE files to build a support module
    inside an epics-containers build
    """
    macro = name.upper() if macro is None else macro
    path = SUPPORT / name if (path is None) else path

    # add or replace the macro for this module in the global RELEASE file
    add_macro(macro, str(path), RELEASE)

    # bring the global release file into this module with a symlink
    local = path / "configure" / "RELEASE.local"
    local.unlink(missing_ok=True)
    local.symlink_to(RELEASE)

    # make sure this module uses RELEASE.local
    verify_release_includes_local(path / "configure")

    do_dependencies()


@support_cli.command()
def add_libs(
    libs: List[str] = typer.Argument(None, help="list of libraries to add"),
) -> None:
    """
    declare the libraries for this support module for inclusion in IOC Makefile
    """
    add_list_to_file(IOC_LIBS, libs)


@support_cli.command()
def add_dbds(
    dbds: List[str] = typer.Argument(None, help="list of dbd files to add"),
) -> None:
    """
    declare the dbd files for this support module for inclusion in IOC Makefile
    """
    add_list_to_file(IOC_DBDS, dbds)


@support_cli.command()
def add_release_macro(
    macro: str = typer.Argument(..., help="macro name to update"),
    value: str = typer.Argument("", help="value to set for the macro"),
    replace: bool = typer.Option(True, help="overwrite previous value"),
):
    """
    add or replace a macro the global RELEASE file
    """
    add_macro(macro, value, RELEASE, replace)


@support_cli.command()
def add_config_macro(
    name: str = typer.Argument(..., help="the name of the support module"),
    macro: str = typer.Argument(..., help="macro name to update"),
    value: str = typer.Argument("", help="value to set for the macro"),
    replace: bool = typer.Option(True, help="overwrite previous value"),
    host: Annotated[Arch, typer.Option(case_sensitive=False)] = Arch.x86_64,
    target: Annotated[Arch, typer.Option(case_sensitive=False)] = Arch.common,
):
    """
    add or replace a macro in CONFIG_SITE.linux-x86_64.Common file
    """
    config_site = get_config_site_file(name, host, target)
    add_macro(macro, value, config_site, replace)


@support_cli.command()
def add_to_config_site(
    module: str = typer.Argument(..., help="support module name"),
    text: str = typer.Argument(..., help="text to add in an idempotent fashion"),
    host: Annotated[Arch, typer.Option(case_sensitive=False)] = Arch.x86_64,
    target: Annotated[Arch, typer.Option(case_sensitive=False)] = Arch.common,
):
    """
    add some text to a support module's CONFIG_SITE file
    """

    # nothing to do if text is blank
    if text != "":
        config_site = get_config_site_file(module, host, target)
        add_text_once(config_site, text)


@support_cli.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def compile(
    ctx: typer.Context,
    module: str = typer.Argument(..., help="support module name"),
):
    """
    compile a support module after preparation with `ibek support register` etc.
    """
    path = SUPPORT / module

    command = f"make -C {path} -j $(nproc) " + " ".join(ctx.args)
    exit(subprocess.call(["bash", "-c", command]))


@support_cli.command()
def generate_links(
    ibek_support: Annotated[
        Optional[Path],
        typer.Option(
            help="Filepath to ibek-support root"
            "defaults to <generic IOC source folder>/ibek-support"
        ),
    ] = None,
):
    """
    generate symlinks to the bob, pvi and support YAML for a compiled IOC
    """

    # symlink the bob YAML
    # TODO TODO
    # symlink the pvi YAML

    # TODO ALSO symlink pvi YAML and read the bobs out of e.g. links/ibek/ADCore.ibek.support.yaml
    # and link them too
    # see line 33 gen_scripts.py

    # symlink the support YAML
    from_path = ibek_support or get_ioc_source() / "ibek-support"

    support_yaml = from_path.glob("*/*.ibek.support.yaml")

    to_path = IBEK_FILES
    to_path.mkdir(parents=True, exist_ok=True)
    for yaml in support_yaml:
        link_from = to_path / yaml.name
        link_from.unlink(missing_ok=True)
        typer.echo(f"symlinking {yaml} to {to_path}")
        link_from.symlink_to(yaml)


@support_cli.command()
def generate_schema(
    output: Annotated[
        Optional[Path], typer.Option(help="The filename to write the schema to")
    ] = None,
):
    """Produce JSON global schema for all <support_module>.ibek.support.yaml files"""
    if output is None:
        typer.echo(Support.get_schema())
    else:
        output.write_text(Support.get_schema())



"""
A class for rendering a substitution file from multiple instantiations of
support module definitions
"""

from dataclasses import dataclass
from typing import Any, Dict, List

from ibek.globals import render_with_utils
from ibek.ioc import IOC, Entity


class RenderDb:
    @dataclass
    class RenderDbTemplate:
        filename: str
        rows: List[List[str]]
        columns: List[int]

    def __init__(self, ioc_instance: IOC) -> None:
        self.ioc_instance = ioc_instance
        # a mapping from template file name to details of instances of that template
        self.render_templates: Dict[str, RenderDb.RenderDbTemplate] = {}

    def add_row(self, filename: str, args: Dict[str, Any], entity: Entity) -> None:
        """
        Accumulate rows of arguments for each template file,
        Adding a new template file if it does not already exist.
        Convert all arguments to strings.
        """
        if filename not in self.render_templates:
            # for new filenames create a new RenderDbTemplate entry
            headings = [str(i) for i in list(args.keys())]
            self.render_templates[filename] = RenderDb.RenderDbTemplate(
                filename=filename,
                rows=[headings],  # first row is the headings
                columns=[0] * len(args),
            )

        # Add a new row of argument values.
        # Note the case where no value was specified for the argument
        # meaning that the name is to be rendered in Jinja
        row = ["{{ %s }}" % k if v is None else v for k, v in args.items()]

        # render any Jinja fields in the arguments
        for i, line in enumerate(row):
            row[i] = render_with_utils(dict(entity), row[i])

        # save the new row
        self.render_templates[filename].rows.append(row)

    def parse_instances(self) -> None:
        """
        Gather the database template instantiations from all entities
        while validating the arguments
        """
        for entity in self.ioc_instance.entities:
            templates = entity.__definition__.databases

            # Not all entities instantiate database templates
            if templates is None or not entity.entity_enabled:
                continue

            for template in templates:
                template.file = template.file.strip("\n")

                for arg, value in template.args.items():
                    if value is None:
                        if arg not in entity.__dict__:
                            raise ValueError(
                                f"database arg '{arg}' in database template "
                                f"'{template.file}' not found in context"
                            )
                self.add_row(template.file, template.args, entity)

    def align_columns(self) -> None:
        """
        Make sure columns will line up for each template file, also
        provide escaping for spaces and quotes
        """

        # first calculate the column width for each template
        # including escaping spaces and quotes
        for template in self.render_templates.values():
            for n, row in enumerate(template.rows):
                for i, arg in enumerate(row):
                    row[i] = f'"{row[i]}"'
                    if i < len(template.columns) - 1:
                        row[i] += ", "
                    template.columns[i] = max(template.columns[i], len(row[i]))

        # now pad each column to the maximum width
        for template in self.render_templates.values():
            for row in template.rows:
                for i, arg in enumerate(row):
                    row[i] = arg.ljust(template.columns[i])

    def render_database(self) -> Dict[str, List[str]]:
        """
        Render a database substitution file
        """
        self.parse_instances()
        self.align_columns()

        results = {}

        for template in self.render_templates.values():
            results[template.filename] = ["".join(row) for row in template.rows]

        return results



"""
Unit tests for the rendering of scripts and database entries from generated
Entity classes
"""
from typing import Literal

from ibek.ioc import IOC, clear_entity_model_ids
from ibek.render import Render
from ibek.render_db import RenderDb


def find_entity_class(entity_classes, entity_type):
    for entity_class in entity_classes:
        literal = Literal[entity_type]  # type: ignore
        if entity_class.model_fields["type"].annotation == literal:
            return entity_class
    else:
        raise ValueError(f"{entity_type} not found in entity_classes")


def test_pre_init_script(objects_classes):
    ref_cls = find_entity_class(objects_classes, "object_module.RefObject")

    my_ref = ref_cls(name="test_ref_object")

    render = Render()
    script_txt = render.render_script(my_ref, my_ref.__definition__.pre_init)
    assert script_txt == (
        "# This line should appear once only in the pre_init section\n"
        "# TestValues testValue\n"
        "TestValues test_ref_object.127.0.0.1\n"
    )


def test_obj_ref_script(objects_classes):
    ref_obj = find_entity_class(objects_classes, "object_module.RefObject")
    consumer = find_entity_class(objects_classes, "object_module.ConsumerTwo")

    ref_obj(name="test_ref_object")
    my_consumer = consumer(name="test_consumer", PORT="test_ref_object")

    render = Render()
    script_txt = render.render_script(my_consumer, my_consumer.__definition__.pre_init)

    assert (
        script_txt == "# ExampleTestFunction asynPortIP name port value\n"
        "ExampleTestFunction 127.0.0.1 test_consumer test_ref_object "
        "test_ref_object.127.0.0.1\n"
        "# PORT defaults to the id of PORT, i.e. PORT.name\n"
    )


def test_database_render(objects_classes, templates):
    ref_cls = find_entity_class(objects_classes, "object_module.RefObject")
    consumer = find_entity_class(objects_classes, "object_module.ConsumerTwo")

    ref_cls(name="test_ref_object")
    my_consumer = consumer(name="test_consumer", PORT="test_ref_object")

    # make a dummy IOC with two entities as database render works against
    # a whole IOC rather than a single entity at a time.
    clear_entity_model_ids()

    ioc = IOC(
        ioc_name="test_ioc",
        description="for testing",
        entities=[],
    )

    ioc.entities.append(my_consumer)
    render_db = RenderDb(ioc)
    templates = render_db.render_database()

    assert templates == {
        "test.db": [
            '"name",          "ip",        "value"                    ',
            '"test_consumer", "127.0.0.1", "test_ref_object.127.0.0.1"',
        ]
    }


def test_environment_variables(objects_classes):
    ref_cls = find_entity_class(objects_classes, "object_module.RefObject")

    ref_obj = ref_cls(name="test_ref_object")

    render = Render()
    env_text = render.render_environment_variables(ref_obj)

    assert env_text == "epicsEnvSet REF_OBJECT_NAME test_ref_object\n"


def test_entity_disabled_does_not_render_elements(objects_classes):
    # covered by tests/test_cli.py::test_build_startup_multiple
    # see disabled entities in tests/samples/yaml/objects.ibek.ioc.yaml
    pass



"""
Functions for inserting snippets into files in an idempotent fashion
"""

from enum import Enum
from pathlib import Path
from typing import List

from ibek.globals import SUPPORT


class Local(str, Enum):
    non_local = ""
    local = ".local"


class Arch(str, Enum):
    none = ""
    common = ".Common"
    x86_64 = ".linux-x86_64"
    arm64 = ".linux-arm64"
    vxWorks = ".vxWorks-ppc604_long"
    rtems = ".RTEMS-beatnik"


def get_config_site_file(
    module: str,
    host: Arch = Arch.x86_64,
    target: Arch = Arch.common,
) -> Path:
    """
    Return the path to an CONFIG_SITE file for a support module.
    Default is /epics/module/configure/CONFIG_SITE.local.linux-x86_64.Common
    """
    name = f"CONFIG_SITE{host.value}{target.value}"
    filepath = SUPPORT / module / "configure" / name

    return filepath


def get_release_file(
    module: str,
    host: Arch = Arch.none,
    target: Arch = Arch.none,
    local: Local = Local.local,
) -> Path:
    """
    Return the path to an EPICS RELEASE file for a support
    default is /epics/module/configure/RELEASE.local
    """
    name = f"RELEASE{host.value}{target.value}{local.value}"
    filepath = SUPPORT / module / "configure" / name

    return filepath


def add_list_to_file(file: Path, text_list: List[str]):
    """
    add a sequence of strings into a file, leaving previously existing
    strings where they are
    """
    if len(text_list) == 0:
        return

    for line in text_list:
        # skip blanks and comments
        if line.startswith("#") or line == "":
            continue
        add_text_once(file, line)


def add_text_once(file: Path, text: str):
    """
    Idempotent add of text to a file.
    Creates the file if it doesn't exist.
    """

    if not file.exists():
        file.parent.mkdir(parents=True, exist_ok=True)
        file.write_text(text + "\n")
    else:
        current = file.read_text()
        if text not in current:
            file.write_text(current + text + "\n")



from pathlib import Path
from typing import List

import typer

from ibek.gen_scripts import (
    create_boot_script,
    create_db_script,
    ioc_deserialize,
)
from ibek.globals import NaturalOrderGroup

runtime_cli = typer.Typer(cls=NaturalOrderGroup)


@runtime_cli.command()
def generate(
    instance: Path = typer.Argument(
        ..., help="The filepath to the ioc instance entity file"
    ),
    definitions: List[Path] = typer.Argument(
        ..., help="The filepath to a support module definition file"
    ),
    out: Path = typer.Option(
        default="config/st.cmd",
        help="Path to output startup script",
    ),
    db_out: Path = typer.Option(
        default="config/db.subst",
        help="Path to output database expansion shell script",
    ),
):
    """
    Build a startup script for an IOC instance
    """

    ## TODO TODO
    # here we want add generation of bob files from PVI files
    # and also make a bob index file of buttons
    #
    # you can access the 'opi' definition like this:
    # ioc_instance.entities[0].__definition__.opis()

    ioc_instance = ioc_deserialize(instance, definitions)

    script_txt = create_boot_script(ioc_instance)

    out.parent.mkdir(parents=True, exist_ok=True)

    with out.open("w") as stream:
        stream.write(script_txt)

    db_txt = create_db_script(ioc_instance)

    with db_out.open("w") as stream:
        stream.write(db_txt)



"""
System tests that run the CLI commands and compare the output to expected
results.
"""
import json
import subprocess
import sys
from pathlib import Path

from ibek import __version__
from ibek.ioc import clear_entity_model_ids
from tests.conftest import run_cli


def test_cli_version():
    cmd = [sys.executable, "-m", "ibek", "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__


def test_ibek_schema(tmp_path: Path, samples: Path):
    """generate the global ibek schema"""
    schema_path = tmp_path / "schema.json"
    run_cli("support", "generate-schema", "--output", schema_path)
    expected = json.loads(
        (samples / "schemas" / "ibek.support.schema.json").read_text()
    )

    actual = json.loads((schema_path).read_text())
    assert expected == actual


def test_object_schema(tmp_path: Path, samples: Path):
    """generate schema from a simple support module definition yaml"""
    clear_entity_model_ids()

    schema_path = tmp_path / "objects.ibek.support.schema.json"
    yaml_path = samples / "yaml" / "objects.ibek.support.yaml"
    run_cli("ioc", "generate-schema", yaml_path, "--output", schema_path)

    expected = json.loads(
        (samples / "schemas" / "objects.ibek.ioc.schema.json").read_text()
    )

    actual = json.loads((schema_path).read_text())
    assert expected == actual


def test_ioc_schema(tmp_path: Path, samples: Path):
    clear_entity_model_ids()
    """generate schema for a container with two support modules"""

    schema_combined = tmp_path / "multiple.ibek.ioc.schema.json"
    yaml_path1 = samples / "yaml" / "objects.ibek.support.yaml"
    yaml_path2 = samples / "yaml" / "all.ibek.support.yaml"
    run_cli(
        "ioc", "generate-schema", yaml_path1, yaml_path2, "--output", schema_combined
    )

    expected = json.loads(
        (samples / "schemas" / "multiple.ibek.ioc.schema.json").read_text()
    )

    actual = json.loads((schema_combined).read_text())
    assert expected == actual


def test_build_runtime_single(tmp_path: Path, samples: Path):
    """
    build an ioc runtime script from an IOC instance entity file
    and a single support module definition file

    Also ensure output directory gets generated if it doesn't pre-exist
    """
    clear_entity_model_ids()
    ioc_yaml = samples / "yaml" / "objects.ibek.ioc.yaml"
    support_yaml = samples / "yaml" / "objects.ibek.support.yaml"
    out_file = tmp_path / "new_dir" / "st.cmd"
    out_db = tmp_path / "new_dir" / "objects.db.subst"

    run_cli(
        "runtime",
        "generate",
        ioc_yaml,
        support_yaml,
        "--out",
        out_file,
        "--db-out",
        out_db,
    )

    example_boot = (samples / "outputs" / "objects.st.cmd").read_text()
    actual_boot = out_file.read_text()
    assert example_boot == actual_boot

    example_db = (samples / "outputs" / "objects.db.subst").read_text()
    actual_db = out_db.read_text()
    assert example_db == actual_db


def test_build_runtime_multiple(tmp_path: Path, samples: Path):
    """
    build an ioc runtime script from an IOC instance entity file
    and multiple support module definition files

    Also verifies database subst file generation for multiple
    entity instantiations.
    """
    clear_entity_model_ids()
    ioc_yaml = samples / "yaml" / "all.ibek.ioc.yaml"
    support_yaml = samples / "yaml" / "objects.ibek.support.yaml"
    support_yaml2 = samples / "yaml" / "all.ibek.support.yaml"
    out_file = tmp_path / "st.cmd"
    out_db = tmp_path / "all.db.subst"

    run_cli(
        "runtime",
        "generate",
        ioc_yaml,
        support_yaml,
        support_yaml2,
        "--out",
        out_file,
        "--db-out",
        out_db,
    )

    example_boot = (samples / "outputs" / "all.st.cmd").read_text()
    actual_boot = out_file.read_text()
    assert example_boot == actual_boot

    example_db = (samples / "outputs" / "all.db.subst").read_text()
    actual_db = out_db.read_text()
    assert example_db == actual_db


def test_build_utils_features(tmp_path: Path, samples: Path):
    """
    build an ioc runtime script to verify utils features
    """
    clear_entity_model_ids()
    ioc_yaml = samples / "yaml" / "utils.ibek.ioc.yaml"
    support_yaml = samples / "yaml" / "utils.ibek.support.yaml"
    out_file = tmp_path / "st.cmd"
    out_db = tmp_path / "db.subst"

    run_cli(
        "runtime",
        "generate",
        ioc_yaml,
        support_yaml,
        "--out",
        out_file,
        "--db-out",
        out_db,
    )

    example_boot = (samples / "outputs" / "utils.st.cmd").read_text()
    actual_boot = out_file.read_text()
    assert example_boot == actual_boot

    example_db = (samples / "outputs" / "utils.db.subst").read_text()
    actual_db = out_db.read_text()
    assert example_db == actual_db



import json
import shutil
import subprocess
from pathlib import Path
from typing import Annotated, List, Optional

import typer
from jinja2 import Template

from ibek.gen_scripts import ioc_create_model
from ibek.globals import (
    IOC_DBDS,
    IOC_FOLDER,
    IOC_LIBS,
    TEMPLATES,
    NaturalOrderGroup,
)
from ibek.ioc_cmds.docker import build_dockerfile

from .assets import extract_assets, get_ioc_source

ioc_cli = typer.Typer(cls=NaturalOrderGroup)


@ioc_cli.command()
def build_docker(
    start: int = typer.Option(1, help="The step to start at in the Dockerfile"),
    stop: int = typer.Option(999, help="The step to stop at in the Dockerfile"),
    dockerfile: Annotated[
        Path, typer.Option(help="The filepath to the Dockerfile to build")
    ] = Path.cwd()
    / "Dockerfile",
):
    """
    EXPERIMENTAL: Attempt to interpret the Dockerfile and run it's commands
    inside the devcontainer. For internal, incremental builds of the Dockerfile.

    Useful for debugging the Dockerfile without having to build the whole
    container from outside of the IOC devcontainer.
    """
    build_dockerfile(dockerfile, start, stop)


@ioc_cli.command()
def generate_schema(
    definitions: List[Path] = typer.Argument(
        ...,
        help="File paths to one or more support module YAML files",
    ),
    output: Annotated[
        Optional[Path],
        typer.Option(help="The file path to the schema file to be written"),
    ] = None,
):
    """
    Create a json schema from a number of support_module.ibek.support.yaml
    files
    """

    ioc_model = ioc_create_model(definitions)
    schema = json.dumps(ioc_model.model_json_schema(), indent=2)
    if output is None:
        typer.echo(schema)
    else:
        output.write_text(schema)


@ioc_cli.command()
def extract_runtime_assets(
    destination: Path = typer.Argument(
        ...,
        help="The root folder to extract assets into",
    ),
    source: Path = typer.Option(
        Path("/epics"),
        help="The root folder to extract assets from",
    ),
    extras: List[Path] = typer.Option(None, help="list of files to also extract"),
    defaults: bool = typer.Option(True, help="copy the default assets"),
):
    """
    Find all the runtime assets in an EPICS installation and copy them to a
    new folder hierarchy for packaging into a container runtime stage.

    This should be performed in a throw away container stage (runtime_prep)
    as it is destructive of the source folder, because it uses move for speed.
    """
    extract_assets(destination, source, extras, defaults)


@ioc_cli.command()
def make_source_template(
    destination: Annotated[
        Optional[Path],
        typer.Option(
            help="Where to make the ioc folder. Defaults to under the "
            "generic IOC source folder",
        ),
    ] = None
):
    """
    Create a new IOC boilerplate source tree in the given folder.
    Default is ioc source repo / ioc .
    """
    if destination is None:
        destination = get_ioc_source() / "ioc"

    if not destination.exists():
        if destination.is_symlink():
            destination.unlink()
        shutil.copytree(TEMPLATES / "ioc", destination)

    # make a symlink to the ioc folder in the root of the epics folder
    # this becomes the standard place for code to look for the IOC
    try:
        IOC_FOLDER.unlink(missing_ok=True)
    except IsADirectoryError:
        shutil.rmtree(IOC_FOLDER)
    IOC_FOLDER.symlink_to(destination)


@ioc_cli.command()
def generate_makefile(
    folder_override: Annotated[
        Optional[Path],
        typer.Option(
            help="Where to find the Makefile.jinja template.",
        ),
    ] = None,
):
    """
    get the dbd and lib files from all support modules and generate
    iocApp/src/Makefile from iocApp/src/Makefile.jinja
    """
    # Folder containing Makefile.jinja
    make_folder = folder_override or get_ioc_source() / "ioc" / "iocApp" / "src"

    # get all the dbd and lib files gathered from each support module
    dbds: List[str] = []
    libs: List[str] = []
    if IOC_DBDS.exists():
        dbds = [dbd.strip() for dbd in IOC_DBDS.read_text().split()]
    if IOC_LIBS.exists():
        libs = [lib.strip() for lib in IOC_LIBS.read_text().split()]

    # generate the Makefile from the template
    template = Template((make_folder / "Makefile.jinja").read_text())
    # libraries are listed in reverse order of dependency
    libs.reverse()
    text = template.render(dbds=dbds, libs=libs)

    with (make_folder / "Makefile").open("w") as stream:
        stream.write(text)


@ioc_cli.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def compile(
    ctx: typer.Context,
):
    """
    Compile a generic IOC after support modules are registered and compiled
    """
    path = IOC_FOLDER

    command = f"make -C {path} -j $(nproc) " + " ".join(ctx.args)
    exit(subprocess.call(["bash", "-c", command]))


@ioc_cli.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def build(
    ctx: typer.Context,
):
    """
    A convenience function that calls make-source-template, generate-makefile,
    compile
    """
    make_source_template()
    generate_makefile()
    compile(ctx)



"""
Tests to verify that the error handling works as expected.
"""
from pathlib import Path

import pytest
from ruamel.yaml import YAML

from ibek.ioc import clear_entity_model_ids, make_entity_models, make_ioc_model
from ibek.support import Support
from ibek.utils import UTILS
from tests.conftest import run_cli


def test_counter_reuse(tmp_path: Path, samples: Path):
    """
    Check you cannot redefine a counter with the same name and different params
    """
    UTILS.__reset__()

    entity_file = samples / "yaml" / "bad_counter.ibek.ioc.yaml"
    definition_file1 = samples / "yaml" / "utils.ibek.support.yaml"

    with pytest.raises(ValueError) as ctx:
        run_cli("runtime", "generate", entity_file, definition_file1)
    assert (
        str(ctx.value)
        == "Redefining counter InterruptVector with different start/stop values"
    )


def test_counter_overuse(tmp_path: Path, samples: Path):
    """
    Check that counter limits are enforced
    """
    UTILS.__reset__()

    entity_file = samples / "yaml" / "bad_counter2.ibek.ioc.yaml"
    definition_file1 = samples / "yaml" / "utils.ibek.support.yaml"

    with pytest.raises(ValueError) as ctx:
        run_cli("runtime", "generate", entity_file, definition_file1)
    assert str(ctx.value) == "Counter 195 exceeded stop value of 194"


def test_bad_ref(tmp_path: Path, samples: Path):
    """
    Check bad object references are caught
    """
    UTILS.__reset__()

    entity_file = samples / "yaml" / "bad_ref.ibek.ioc.yaml"
    definition_file1 = samples / "yaml" / "objects.ibek.support.yaml"
    definition_file2 = samples / "yaml" / "all.ibek.support.yaml"

    with pytest.raises(ValueError) as ctx:
        run_cli("runtime", "generate", entity_file, definition_file1, definition_file2)
    assert "object Ref2 not found" in str(ctx.value)


def test_bad_db(tmp_path: Path, samples: Path):
    """
    Check bad database args are caught
    """
    UTILS.__reset__()

    entity_file = samples / "yaml" / "bad_db.ibek.ioc.yaml"
    definition_file1 = samples / "yaml" / "bad_db.ibek.support.yaml"

    with pytest.raises(ValueError) as ctx:
        run_cli("runtime", "generate", entity_file, definition_file1)
    assert "'non-existant' in database template 'test.db' not found" in str(ctx.value)


def test_loading_module_twice(tmp_path: Path, samples: Path):
    """
    Verify we get a sensible error if we try to load a module twice
    without clearing the entity model ids
    """

    clear_entity_model_ids()

    definition_file = samples / "yaml" / "objects.ibek.support.yaml"
    instance_file = samples / "yaml" / "objects.ibek.ioc.yaml"

    support = Support(**YAML(typ="safe").load(definition_file))
    entities1 = make_entity_models(support)
    entities2 = make_entity_models(support)

    generic_ioc1 = make_ioc_model(entities1)
    generic_ioc2 = make_ioc_model(entities2)

    instance = YAML(typ="safe").load(instance_file)
    generic_ioc1(**instance)
    with pytest.raises(ValueError) as ctx:
        generic_ioc2(**instance)

    assert "Duplicate id" in str(ctx.value)


def test_defaults(tmp_path: Path, samples: Path):
    """
    Check you cannot redefine a counter with the same name and different params
    """

    clear_entity_model_ids()
    entity_file = samples / "yaml" / "bad_default.ibek.ioc.yaml"
    definition_file1 = samples / "yaml" / "objects.ibek.support.yaml"

    with pytest.raises(ValueError) as ctx:
        run_cli("runtime", "generate", entity_file, definition_file1)

    assert "Field required [type=missing" in str(ctx.value)
